clear
close all
clc

%% 1. Design a digital highpass filter
% Second-order IIR
% Cutoff frequency fc = 50 Hz
% Butterworth design method
f_S         = 44100;
f_c         = 50;
N           = 2;
NFFT        = 4096;
f_vec       = (0:NFFT-1) * f_S/NFFT;
x           = zeros(NFFT, 1);
x(1)        = 1;

[b, a]      = butter(N, f_c/(f_S/2), 'high');
h_1         = filter(b, a, x);

figure;
semilogx(f_vec, 20*log10(abs(fft(h_1,NFFT))), 'k');
xlim([0 f_S/2]); ylim([-50 30]); grid on;
xlabel('Frequency in Hz'); ylabel('Magnitude in dB');
title('Digital highpass filter');

%% 2. Quantize signal when it leaves the accumulator
w_y         = 28; 
y_max       = 1;
Q_y         = (2*y_max)/(2^w_y);
y           = zeros(NFFT, 1);
x_hist      = zeros(N, 1);
y_hist      = zeros(N, 1);
for n = 1 : NFFT
    % Accumulator
    y(n)    = b(1) * x(n);
    for m = 2 : length(b)
        y(n) 	= y(n) + b(m) * x_hist(m-1);
    end
    for k = 2 : length(a)
        y(n)  	= y(n) - a(k) * y_hist(k-1);
    end
    % Quantization
    y(n)        = Q_y * floor(y(n)/Q_y+0.5);
    % Storing previous inputs and outputs
    x_hist      = [x(n); x_hist(1:end-1,1)]; 
    y_hist      = [y(n); y_hist(1:end-1,1)];
    clear m k
end
h_2         = y;
clear y x_hist y_hist 

figure;
semilogx(f_vec, 20*log10(abs(fft(h_2,NFFT))), 'b');
xlim([0 f_S/2]); ylim([-50 30]); grid on;
xlabel('Frequency in Hz'); ylabel('Magnitude in dB');
title('Quantization after accumulator');

%% 3. Quantize coefficients (direct form) too
y           = zeros(NFFT, 1);
x_hist      = zeros(N, 1);
y_hist      = zeros(N, 1);
% Quantization of coefficients
w_c         = 18; 
c_max       = 2;
Q_c         = (2*c_max)/(2^w_c);
b_c         = Q_c * floor(b/Q_c+0.5);
a_c         = Q_c * floor(a/Q_c+0.5);
for n = 1 : NFFT
    % Accumulator
    y(n)    = b_c(1) * x(n);
    for m = 2 : length(b)
        y(n) 	= y(n) + b_c(m) * x_hist(m-1);
    end
    for k = 2 : length(a)
        y(n)  	= y(n) - a_c(k) * y_hist(k-1);
    end
    % Quantization
    y(n)        = Q_y * floor(y(n)/Q_y+0.5);
    % Storing previous inputs and outputs
    x_hist      = [x(n); x_hist(1:end-1,1)]; 
    y_hist      = [y(n); y_hist(1:end-1,1)];
    clear m k
end
h_3         = y;
clear y x_hist y_hist 

figure;
semilogx(f_vec, 20*log10(abs(fft(h_3,NFFT))), 'r');
xlim([0 f_S/2]); ylim([-50 30]); grid on;
xlabel('Frequency in Hz'); ylabel('Magnitude in dB');
title('Quantization after accumulator + quantized coefficients');

%% 4. Quantization after every arithmetic operation
y           = zeros(NFFT, 1);
x_hist      = zeros(N, 1);
y_hist      = zeros(N, 1);
for n = 1 : NFFT
    % Accumulator
    y(n)    = b_c(1) * x(n);
    y(n)    = Q_y * floor(y(n)/Q_y+0.5);
    for m = 2 : length(b)
        y(n) 	= y(n) + b_c(m) * x_hist(m-1);
        y(n)    = Q_y * floor(y(n)/Q_y+0.5);
    end
    for k = 2 : length(a)
        y(n)  	= y(n) - a_c(k) * y_hist(k-1);
        y(n)    = Q_y * floor(y(n)/Q_y+0.5);
    end
    % Storing previous inputs and outputs
    x_hist      = [x(n); x_hist(1:end-1,1)]; 
    y_hist      = [y(n); y_hist(1:end-1,1)];
    clear m k
end
h_4         = y;
clear y x_hist y_hist 

figure;
semilogx(f_vec, 20*log10(abs(fft(h_4,NFFT))), 'm');
xlim([0 f_S/2]); ylim([-50 30]); grid on;
xlabel('Frequency in Hz'); ylabel('Magnitude in dB');
title('Quantization after every arithmetic operation');

%% Comparison

figure;
semilogx(f_vec, 20*log10(abs(fft(h_1,NFFT))), 'k'); hold on;
semilogx(f_vec, 20*log10(abs(fft(h_2,NFFT))), 'b');
semilogx(f_vec, 20*log10(abs(fft(h_3,NFFT))), 'r');
semilogx(f_vec, 20*log10(abs(fft(h_4,NFFT))), 'm'); hold off;
xlim([0 f_S/2]); ylim([-50 30]); grid on;
xlabel('Frequency in Hz'); ylabel('Magnitude in dB');
legend({'without quantization','quantization after accumulator','+ coefficient quantization','quantization after every aithmetic operation'},'Location','Northeast');
